==> luna_cow-0.5.79-py3-none-any.whl: cow/web/application.py <==
"""
LunaApplication is an aggregate of objects required to start and configure service.
It wraps sanic application, effectively separating namespaces.

Attributes that are meant to be initialized and accessed during runtime should be
set in context class.
"""
import asyncio
import dataclasses
import platform
import signal
from itertools import chain
from pathlib import Path
from typing import Awaitable, Generic, Type

import log
from luna_plugins.base.manager import PluginManager
from sanic import Sanic as BaseSanic
from sanic.handlers import ErrorHandler
from sanic.mixins.listeners import ListenerEvent
from sanic.views import HTTPMethodView
from sanic.worker.loader import AppLoader

from ..constants.http_consts import (
    STANDARD_KEEP_ALIVE_TIMEOUT,
    STANDARD_REQUEST_MAX_SIZE,
    STANDARD_REQUEST_TIMEOUT,
    STANDARD_RESPONSE_TIMEOUT,
)
from ..constants.paths import APP_ROOT
from ..constants.third_party_loggers import THIRD_PARTY_LOGGERS
from ..errors.errors import Error
from ..errors.exception import VLException
from ..monitoring.base_monitoring import LunaRequestInfluxMonitoring
from ..utils import mixins
from ..utils.healthcheck import Health, performCheck
from ..utils.log import logger
from .cmd_parser import ApplicationCMDParser
from .context import CONFIG_CLASS, CONTEXT_CLASS
from .error_handlers import serverErrorHandler
from .middlewares import (
    compressResponseMiddleware,
    deCompressResponseMiddleware,
    endRequestMiddleware,
    logErrorMiddleware,
    restMiddleware,
    sendToMonitoringMiddleware,
)
from .request import VLRequest


@dataclasses.dataclass
class BaseLunaApplicationCtx(Generic[CONFIG_CLASS]):
    """Base luna application context. Legacy"""

    #: service api version
    apiVersion: int | None = None
    #: default port
    defaultPort: int | None = None
    #: service host
    host: str | None = None
    #: service port
    port: int | None = None
    #: service monitoring
    monitoring: LunaRequestInfluxMonitoring | None = None


class LunaApplication(Generic[CONFIG_CLASS, CONTEXT_CLASS]):
    """
    Proxy class for new LunaApp
    """

    def __init__(
        self,
        apiVersion: int,
        configClass: Type[CONFIG_CLASS],
        defaultPort: int,
        ctxClass: Type[CONTEXT_CLASS],
        requestClass: Type[VLRequest] = VLRequest,
        **kwargs,  # pylint: disable-msg=W0613
    ):
        LunaApp.configCls = configClass
        LunaApp.defaultPort = defaultPort
        LunaApp.requestCls = requestClass
        LunaApp.ctxCls = ctxClass
        LunaApp.apiVersion = apiVersion

        self.ctx: CONTEXT_CLASS = ctxClass()
        self.apiVersion: int = apiVersion

        if platform.system() != "Windows":
            import uvloop  # pylint: disable-msg=import-outside-toplevel

            uvloop.install()

        cmdArgs = self._getCMDArgParser(defaultPort=defaultPort, serviceConfig=configClass()).parse()  # handlers

        self._app: LunaApp[CONFIG_CLASS, CONTEXT_CLASS] = LunaApp(legacyCtx=self.ctx, legacyCmdArgs=cmdArgs)

        self.sanic = self._app.sanic
        self.addRoutes = self._app.addRoutes
        self.register_middleware = self.sanic.register_middleware
        self.add_route = self.sanic.add_route
        self.add_websocket_route = self.sanic.add_websocket_route
        self.response_middleware = self.sanic.response_middleware
        self.url_for = self.sanic.url_for  # admin
        self.middleware = self.sanic.middleware  # admin

        # configurator
        self.static = self.sanic.static

        # lpm
        self.config = self.sanic.config
        self.register_listener = self.sanic.register_listener

        self.run = self._app.run
        self.ctx.getRuntimeChecks = self.getRuntimeChecks

        self.ctx.apiVersion = apiVersion
        self.ctx.cmdArgs = cmdArgs
        self.ctx.serviceConfig = self._app.serviceConfig
        self.ctx.logger = logger
        self.name = self.ctx.serviceConfig.serviceName

        async def initializeApp(*_):
            await self.initialize()

        async def shutdown(*_):
            logger.info("Stopping the application")
            await self.ctx.pluginManager.shutdown()

            await self.shutdown()

            if self.ctx.monitoring:
                await self.ctx.monitoring.close()

            logger.info("Application has stopped")

        self.sanic.register_listener(initializeApp, ListenerEvent.BEFORE_SERVER_START.name)
        self.sanic.register_listener(shutdown, ListenerEvent.AFTER_SERVER_STOP.name)

    async def initialize(self):
        """
        Initialize service
        """
        self.ctx.monitoring = LunaRequestInfluxMonitoring(
            host=self.ctx.serviceConfig.monitoring.host,
            port=self.ctx.serviceConfig.monitoring.port,
            credentials=self.ctx.serviceConfig.monitoring.credentials,
            ssl=self.ctx.serviceConfig.monitoring.useSsl,
            flushingPeriod=self.ctx.serviceConfig.monitoring.flushingPeriod,
        )
        if self.ctx.serviceConfig.monitoring.sendData:
            await self.ctx.monitoring.initialize()

        self.ctx.pluginManager = PluginManager(self, self.pluginDirectory)
        await self.ctx.pluginManager.initialize(self.ctx.serviceConfig.activePlugins)
        await self.ctx.pluginManager.start()

    async def shutdown(self):
        """
        Shutdown service

        Override this in service subclasses.
        """

    def getRuntimeChecks(self, includeLunaServices: bool = False) -> list[tuple[str, Awaitable[None]]]:
        # pylint: disable-msg=unused-argument
        """
        Returns configured system checks, pairs of (name, coroutine function).

        Override this in service subclasses.
        """
        return []

    async def checkHealth(self, includeLunaServices: bool = False) -> list[Health]:
        """
        Run checks simultaneously, return list of checks results.

        Args:
            includeLunaServices: A bool, whether to return checks for luna services.
        """
        waiters = [performCheck(name, check) for name, check in self.getRuntimeChecks(includeLunaServices)]
        results = await asyncio.gather(*waiters)
        return sorted(list(chain(*results)), key=lambda check: check.status)

    @staticmethod
    def _getCMDArgParser(defaultPort: int, serviceConfig: CONFIG_CLASS) -> ApplicationCMDParser:
        """
        Get command line arguments parser
        Returns:
            parser with standard  options
        """
        argParser = ApplicationCMDParser(
            defaultPort=defaultPort, tagableSettings=serviceConfig.getServiceSettingNames()
        )
        return argParser

    @property
    def pluginDirectory(self) -> Path:
        """Get plugin directory"""
        return APP_ROOT / "plugins"


class LunaApp(Generic[CONFIG_CLASS, CONTEXT_CLASS]):
    """Luna service web application"""

    ctxCls: Type[CONTEXT_CLASS]
    configCls: Type[CONFIG_CLASS]
    defaultPort: int
    apiVersion: int
    requestCls = VLRequest

    def __init__(
        self,
        thirdPartyLoggerWhiteList: list[str] | None = None,
        serviceConfig: CONFIG_CLASS | None = None,
        defaultPort: int | None = None,
        legacyCtx: CONTEXT_CLASS | None = None,  # COMPAT
        legacyCmdArgs=None,  # COMPAT
    ):
        """
        Args:
            legacyCtx: initialized context
        """
        if not thirdPartyLoggerWhiteList:
            thirdPartyLoggerWhiteList = THIRD_PARTY_LOGGERS
        else:
            thirdPartyLoggerWhiteList.extend(THIRD_PARTY_LOGGERS)

        if defaultPort is not None:
            self.defaultPort = defaultPort

        if not legacyCmdArgs:
            self.cmdArgs = cmdArgs = self._getCMDArgParser().parse()
        else:
            self.cmdArgs = cmdArgs = legacyCmdArgs

        if not serviceConfig:
            serviceConfig = self.loadConfig()
        self.serviceConfig = serviceConfig
        self.configureLogger(cmdArgs.logSuffix, thirdPartyLoggerWhiteList)

        # compat
        self.name = serviceConfig.serviceName

        sanic = self.sanic = Sanic(
            name=self.name,
            strict_slashes=False,
            request_class=self.requestCls,
            configure_logging=False,
            error_handler=LunaErrorHandler(),
        )
        self.setUpSanic()

        pluginManager = PluginManager(self, APP_ROOT / "plugins")

        async def _initializeCtx(*_):
            """
            Initialze components. Order matters.
            Pluginmanager always in the end.
            """
            # using this hack to prevent race between sanic listeners (old and new one)
            if legacyCtx:
                sanic.ctx = legacyCtx
                return

            sanic.ctx = ctx = self.ctxCls(
                serviceConfig, pluginManager
            )  # pylint: disable-msg=attribute-defined-outside-init
            ctx.cmdArgs = cmdArgs

            try:
                await ctx.initialize()
            except mixins.InitializationError:
                self.sanic.stop()

            # plugins have inclusive access to ctx, initializing them after others
            await pluginManager.initialize(serviceConfig.activePlugins)
            await pluginManager.start()

        sanic.register_listener(_initializeCtx, ListenerEvent.BEFORE_SERVER_START.name)

        async def _shutdown(*_):
            """
            Shutdown components from leafs to roots.
            """
            # COMPAT
            if legacyCtx:
                return

            await pluginManager.shutdown()
            await sanic.ctx.close()

        sanic.register_listener(_shutdown, ListenerEvent.AFTER_SERVER_STOP.name)

        async def _printServiceStart(app, loop):  # pylint: disable-msg=W0613
            """
            Print a message about that service is started
            """
            host, port = sanic.state.host, sanic.state.port
            host = host if host != "0.0.0.0" else "127.0.0.1"
            print(f"======== Running on http://{host}:{port} ========")
            logger.info(f"Service is running on http://{host}:{port}")

        sanic.register_listener(_printServiceStart, ListenerEvent.AFTER_SERVER_START.name)

    def setUpSanic(self):
        """Set up sanic"""
        sanic = self.sanic
        sanic.register_middleware(deCompressResponseMiddleware, "request")
        sanic.register_middleware(endRequestMiddleware, "response")
        sanic.register_middleware(restMiddleware, "response")
        sanic.register_middleware(compressResponseMiddleware, "response")
        sanic.register_middleware(sendToMonitoringMiddleware, "response")
        sanic.register_middleware(logErrorMiddleware, "response")

        # see https://sanic.readthedocs.io/en/latest/sanic/config.html#builtin-configuration-values for details
        if httpSettings := getattr(self.serviceConfig, "httpSettings", None):
            requestTimeout = httpSettings.requestTimeout
            responseTimeout = httpSettings.responseTimeout
            requestMaxSize = httpSettings.requestMaxSize
            keepAliveTimeout = httpSettings.keepAliveTimeout
        else:
            requestTimeout = STANDARD_REQUEST_TIMEOUT
            responseTimeout = STANDARD_RESPONSE_TIMEOUT
            requestMaxSize = STANDARD_REQUEST_MAX_SIZE
            keepAliveTimeout = STANDARD_KEEP_ALIVE_TIMEOUT
        sanic.config["REQUEST_TIMEOUT"] = requestTimeout
        sanic.config["RESPONSE_TIMEOUT"] = responseTimeout
        sanic.config["REQUEST_MAX_SIZE"] = requestMaxSize
        sanic.config["KEEP_ALIVE_TIMEOUT"] = keepAliveTimeout
        sanic.config["WEBSOCKET_PING_INTERVAL"] = None
        sanic.config["ACCESS_LOG"] = False

    @property
    def ctx(self) -> CONTEXT_CLASS:
        """Get current service context"""
        return self.sanic.ctx

    def _getCMDArgParser(self) -> ApplicationCMDParser:
        """
        Get command line arguments parser
        Returns:
            parser with standard  options
        """
        argParser = ApplicationCMDParser(self.defaultPort, tagableSettings=self.configCls.getServiceSettingNames())
        return argParser

    def loadConfig(self) -> CONFIG_CLASS:
        """Fetch and apply configuration."""
        serviceConfig = self.configCls()
        serviceConfig.load(
            configFile=self.cmdArgs.config,
            configuratorAddress=self.cmdArgs.configuratorAddress,
            tags=self.cmdArgs.tags,
            asyncLoad=False,
        )
        return serviceConfig

    def configureLogger(self, logSuffix: str, thirdPartyLoggerWhiteList: list[str]):
        """Given cmdArgs and serviceConfig, perform logging configuraiton."""
        config = self.serviceConfig
        log.initiate(
            appName=config.serviceName,
            logLevel=config.logger.logLevel,
            logTime=config.logger.logTime,
            folderForLog=config.logger.folderWithLogs,
            maxSize=config.logger.maxLogFileSize,
            logNameSuffix=logSuffix,
            logToStdOut=config.logger.logToStdOut,
            logToFile=config.logger.logToFile,
            multilineStackTrace=config.logger.multilineStackTrace,
            thirdPartyLoggerWhiteList=thirdPartyLoggerWhiteList,
            formatter=config.logger.format,
        )

    def addRoutes(self, routes: list[tuple[str, Type[HTTPMethodView]]]):
        """
        Add routes.

        This method accepts type or view function as view.

        Args:
            routes: list of routes (<resource, HTTPMethodView>)
        """
        for resource, view in routes:
            if hasattr(view, "view_class"):
                self.sanic.add_route(view, resource)
            else:
                self.sanic.add_route(view.as_view(), resource)

    def run(self, host="0.0.0.0", **kwargs):
        """Run sanic application."""
        app = AppLoader(factory=lambda: self.sanic).load()

        sock = kwargs.pop("sock", None)
        if sock:
            _, port = sock.getsockname()
            app.prepare(host=host, port=port, **kwargs)
            # if sock is specified sanic wont bind a new one
            # otherwise app will get stuck and tests wont pass
            app.state.server_info[0].settings["sock"] = sock
        else:
            port = kwargs.pop("port", None) or self.cmdArgs.port
            app.prepare(host=host, port=port, **kwargs)
        Sanic.serve_single(primary=app)


class LunaErrorHandler(ErrorHandler):
    """Common error handler"""

    def default(self, request: VLRequest, exception):
        """Default error handler."""
        return serverErrorHandler(request, exception)


class Sanic(BaseSanic):  # pylint: disable-msg=too-many-ancestors
    """Sanic customization."""

    async def _static_request_handler(self, *args, **kwargs):  # pylint: disable-msg=C0103,W0221
        """
        Overrides sanic static processing because it returns None on `IsADirectory` exception that leads to 500 error
        """
        result = await super()._static_request_handler(*args, **kwargs)  # pylint: disable-msg=E1101
        if result is None:
            raise VLException(Error.PageNotFoundError, 404, False)
        return result

    async def _websocket_handler(
        self, handler: "WSBaseHandler", request: VLRequest, *args, subprotocols=None, **kwargs
    ):
        """
        Websocket handle. Open ws session with validation and preparation.
        """
        wsSession = getattr(handler, "wsSession", None)  # hack for data validation before handshake
        if wsSession:
            handler = handler(request, wsSession)
            async with handler.wsSession(request):
                await super()._websocket_handler(
                    lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs
                )
        else:
            await super()._websocket_handler(
                lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs
            )

    def stop(self, *args, **kwargs):
        """
        Stops sanic application, taking care of not shutting down twice.
        """
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        super().stop(*args, **kwargs)

==> luna_cow-0.5.86-py3-none-any.whl: cow/web/application.py <==
"""
LunaApplication is an aggregate of objects required to start and configure service.
It wraps sanic application, effectively separating namespaces.

Attributes that are meant to be initialized and accessed during runtime should be
set in context class.
"""
import asyncio
import dataclasses
import platform
import signal
from itertools import chain
from pathlib import Path
from typing import Awaitable, Generic, Type

import log
from luna_plugins.base.manager import PluginManager
from sanic import Sanic as BaseSanic
from sanic.handlers import ErrorHandler
from sanic.mixins.listeners import ListenerEvent
from sanic.views import HTTPMethodView
from sanic.worker.loader import AppLoader

from ..constants.http_consts import (
    STANDARD_KEEP_ALIVE_TIMEOUT,
    STANDARD_REQUEST_MAX_SIZE,
    STANDARD_REQUEST_TIMEOUT,
    STANDARD_RESPONSE_TIMEOUT,
)
from ..constants.paths import APP_ROOT
from ..constants.third_party_loggers import THIRD_PARTY_LOGGERS
from ..errors.errors import Error
from ..errors.exception import VLException
from ..monitoring.manager import LunaMonitoringManager
from ..utils import mixins
from ..utils.healthcheck import Health, performCheck
from ..utils.log import logger
from .cmd_parser import ApplicationCMDParser
from .context import CONFIG_CLASS, CONTEXT_CLASS
from .error_handlers import serverErrorHandler
from .middlewares import (
    compressResponseMiddleware,
    deCompressResponseMiddleware,
    endRequestMiddleware,
    logErrorMiddleware,
    restMiddleware,
    sendToMonitoringMiddleware,
)
from .request import VLRequest


@dataclasses.dataclass
class BaseLunaApplicationCtx(Generic[CONFIG_CLASS]):
    """Base luna application context. Legacy"""

    #: service api version
    apiVersion: int | None = None
    #: default port
    defaultPort: int | None = None
    #: service host
    host: str | None = None
    #: service port
    port: int | None = None
    #: service monitoring
    monitoring: LunaMonitoringManager | None = None


class LunaApplication(Generic[CONFIG_CLASS, CONTEXT_CLASS]):
    """
    Proxy class for new LunaApp
    """

    def __init__(
        self,
        apiVersion: int,
        configClass: Type[CONFIG_CLASS],
        defaultPort: int,
        ctxClass: Type[CONTEXT_CLASS],
        requestClass: Type[VLRequest] = VLRequest,
        **kwargs,  # pylint: disable-msg=W0613
    ):
        LunaApp.configCls = configClass
        LunaApp.defaultPort = defaultPort
        LunaApp.requestCls = requestClass
        LunaApp.ctxCls = ctxClass
        LunaApp.apiVersion = apiVersion

        self.ctx: CONTEXT_CLASS = ctxClass()
        self.apiVersion: int = apiVersion

        if platform.system() != "Windows":
            import uvloop  # pylint: disable-msg=import-outside-toplevel

            uvloop.install()

        cmdArgs = self._getCMDArgParser(defaultPort=defaultPort, serviceConfig=configClass()).parse()  # handlers

        self._app: LunaApp[CONFIG_CLASS, CONTEXT_CLASS] = LunaApp(legacyCtx=self.ctx, legacyCmdArgs=cmdArgs)

        self.sanic = self._app.sanic
        self.addRoutes = self._app.addRoutes
        self.register_middleware = self.sanic.register_middleware
        self.add_route = self.sanic.add_route
        self.add_websocket_route = self.sanic.add_websocket_route
        self.response_middleware = self.sanic.response_middleware
        self.url_for = self.sanic.url_for  # admin
        self.middleware = self.sanic.middleware  # admin

        # configurator
        self.static = self.sanic.static

        # lpm
        self.config = self.sanic.config
        self.register_listener = self.sanic.register_listener

        self.run = self._app.run
        self.ctx.getRuntimeChecks = self.getRuntimeChecks

        self.ctx.apiVersion = apiVersion
        self.ctx.cmdArgs = cmdArgs
        self.ctx.serviceConfig = self._app.serviceConfig
        self.ctx.logger = logger
        self.name = self.ctx.serviceConfig.serviceName

        async def initializeApp(*_):
            await self.initialize()

        async def shutdown(*_):
            logger.info("Stopping the application")
            await self.ctx.pluginManager.shutdown()

            await self.shutdown()

            if self.ctx.monitoring:
                await self.ctx.monitoring.close()

            logger.info("Application has stopped")

        self.sanic.register_listener(initializeApp, ListenerEvent.BEFORE_SERVER_START.name)
        self.sanic.register_listener(shutdown, ListenerEvent.AFTER_SERVER_STOP.name)

    async def initialize(self):
        """
        Initialize service
        """
        self.ctx.pluginManager = PluginManager(self, self.pluginDirectory)
        await self.ctx.pluginManager.initialize(self.ctx.serviceConfig.activePlugins)
        await self.ctx.pluginManager.start()
        self.ctx.monitoring = LunaMonitoringManager(self.ctx.serviceConfig.monitoring, self.ctx.pluginManager)
        await self.ctx.monitoring.initialize()

    async def shutdown(self):
        """
        Shutdown service

        Override this in service subclasses.
        """

    def getRuntimeChecks(self, includeLunaServices: bool = False) -> list[tuple[str, Awaitable[None]]]:
        # pylint: disable-msg=unused-argument
        """
        Returns configured system checks, pairs of (name, coroutine function).

        Override this in service subclasses.
        """
        return []

    async def checkHealth(self, includeLunaServices: bool = False) -> list[Health]:
        """
        Run checks simultaneously, return list of checks results.

        Args:
            includeLunaServices: A bool, whether to return checks for luna services.
        """
        waiters = [performCheck(name, check) for name, check in self.getRuntimeChecks(includeLunaServices)]
        results = await asyncio.gather(*waiters)
        return sorted(list(chain(*results)), key=lambda check: check.status)

    @staticmethod
    def _getCMDArgParser(defaultPort: int, serviceConfig: CONFIG_CLASS) -> ApplicationCMDParser:
        """
        Get command line arguments parser
        Returns:
            parser with standard  options
        """
        argParser = ApplicationCMDParser(
            defaultPort=defaultPort, tagableSettings=serviceConfig.getServiceSettingNames()
        )
        return argParser

    @property
    def pluginDirectory(self) -> Path:
        """Get plugin directory"""
        return APP_ROOT / "plugins"


class LunaApp(Generic[CONFIG_CLASS, CONTEXT_CLASS]):
    """Luna service web application"""

    ctxCls: Type[CONTEXT_CLASS]
    configCls: Type[CONFIG_CLASS]
    defaultPort: int
    apiVersion: int
    requestCls = VLRequest

    def __init__(
        self,
        thirdPartyLoggerWhiteList: list[str] | None = None,
        serviceConfig: CONFIG_CLASS | None = None,
        defaultPort: int | None = None,
        legacyCtx: CONTEXT_CLASS | None = None,  # COMPAT
        legacyCmdArgs=None,  # COMPAT
    ):
        """
        Args:
            legacyCtx: initialized context
        """
        if not thirdPartyLoggerWhiteList:
            thirdPartyLoggerWhiteList = THIRD_PARTY_LOGGERS
        else:
            thirdPartyLoggerWhiteList.extend(THIRD_PARTY_LOGGERS)

        if defaultPort is not None:
            self.defaultPort = defaultPort

        if not legacyCmdArgs:
            self.cmdArgs = cmdArgs = self._getCMDArgParser().parse()
        else:
            self.cmdArgs = cmdArgs = legacyCmdArgs

        if not serviceConfig:
            serviceConfig = self.loadConfig()
        self.serviceConfig = serviceConfig
        self.configureLogger(cmdArgs.logSuffix, thirdPartyLoggerWhiteList)

        # compat
        self.name = serviceConfig.serviceName

        sanic = self.sanic = Sanic(
            name=self.name,
            strict_slashes=False,
            request_class=self.requestCls,
            configure_logging=False,
            error_handler=LunaErrorHandler(),
        )
        self.setUpSanic()

        pluginManager = PluginManager(self, APP_ROOT / "plugins")

        async def _initializeCtx(*_):
            """
            Initialze components. Order matters.
            Pluginmanager always in the end.
            """
            # using this hack to prevent race between sanic listeners (old and new one)
            if legacyCtx:
                sanic.ctx = legacyCtx
                return

            sanic.ctx = ctx = self.ctxCls(
                serviceConfig, pluginManager
            )  # pylint: disable-msg=attribute-defined-outside-init
            ctx.cmdArgs = cmdArgs

            try:
                await ctx.initialize()
            except mixins.InitializationError:
                self.sanic.stop()

            # plugins have inclusive access to ctx, initializing them after others
            await pluginManager.initialize(serviceConfig.activePlugins)
            await pluginManager.start()

        sanic.register_listener(_initializeCtx, ListenerEvent.BEFORE_SERVER_START.name)

        async def _shutdown(*_):
            """
            Shutdown components from leafs to roots.
            """
            # COMPAT
            if legacyCtx:
                return

            await pluginManager.shutdown()
            await sanic.ctx.close()

        sanic.register_listener(_shutdown, ListenerEvent.AFTER_SERVER_STOP.name)

        async def _printServiceStart(app, loop):  # pylint: disable-msg=W0613
            """
            Print a message about that service is started
            """
            host, port = sanic.state.host, sanic.state.port
            host = host if host != "0.0.0.0" else "127.0.0.1"
            print(f"======== Running on http://{host}:{port} ========")
            logger.info(f"Service is running on http://{host}:{port}")

        sanic.register_listener(_printServiceStart, ListenerEvent.AFTER_SERVER_START.name)

    def setUpSanic(self):
        """Set up sanic"""
        sanic = self.sanic
        sanic.register_middleware(deCompressResponseMiddleware, "request")
        sanic.register_middleware(endRequestMiddleware, "response")
        sanic.register_middleware(restMiddleware, "response")
        sanic.register_middleware(compressResponseMiddleware, "response")
        sanic.register_middleware(sendToMonitoringMiddleware, "response")
        sanic.register_middleware(logErrorMiddleware, "response")

        # see https://sanic.readthedocs.io/en/latest/sanic/config.html#builtin-configuration-values for details
        if httpSettings := getattr(self.serviceConfig, "httpSettings", None):
            requestTimeout = httpSettings.requestTimeout
            responseTimeout = httpSettings.responseTimeout
            requestMaxSize = httpSettings.requestMaxSize
            keepAliveTimeout = httpSettings.keepAliveTimeout
        else:
            requestTimeout = STANDARD_REQUEST_TIMEOUT
            responseTimeout = STANDARD_RESPONSE_TIMEOUT
            requestMaxSize = STANDARD_REQUEST_MAX_SIZE
            keepAliveTimeout = STANDARD_KEEP_ALIVE_TIMEOUT
        sanic.config["REQUEST_TIMEOUT"] = requestTimeout
        sanic.config["RESPONSE_TIMEOUT"] = responseTimeout
        sanic.config["REQUEST_MAX_SIZE"] = requestMaxSize
        sanic.config["KEEP_ALIVE_TIMEOUT"] = keepAliveTimeout
        sanic.config["WEBSOCKET_PING_INTERVAL"] = None
        sanic.config["ACCESS_LOG"] = False

    @property
    def ctx(self) -> CONTEXT_CLASS:
        """Get current service context"""
        return self.sanic.ctx

    def _getCMDArgParser(self) -> ApplicationCMDParser:
        """
        Get command line arguments parser
        Returns:
            parser with standard  options
        """
        argParser = ApplicationCMDParser(self.defaultPort, tagableSettings=self.configCls.getServiceSettingNames())
        return argParser

    def loadConfig(self) -> CONFIG_CLASS:
        """Fetch and apply configuration."""
        serviceConfig = self.configCls()
        serviceConfig.load(
            configFile=self.cmdArgs.config,
            configuratorAddress=self.cmdArgs.configuratorAddress,
            tags=self.cmdArgs.tags,
            asyncLoad=False,
        )
        return serviceConfig

    def configureLogger(self, logSuffix: str, thirdPartyLoggerWhiteList: list[str]):
        """Given cmdArgs and serviceConfig, perform logging configuraiton."""
        config = self.serviceConfig
        log.initiate(
            appName=config.serviceName,
            logLevel=config.logger.logLevel,
            logTime=config.logger.logTime,
            folderForLog=config.logger.folderWithLogs,
            maxSize=config.logger.maxLogFileSize,
            logNameSuffix=logSuffix,
            logToStdOut=config.logger.logToStdOut,
            logToFile=config.logger.logToFile,
            multilineStackTrace=config.logger.multilineStackTrace,
            thirdPartyLoggerWhiteList=thirdPartyLoggerWhiteList,
            formatter=config.logger.format,
        )

    def addRoutes(self, routes: list[tuple[str, Type[HTTPMethodView]]]):
        """
        Add routes.

        This method accepts type or view function as view.

        Args:
            routes: list of routes (<resource, HTTPMethodView>)
        """
        for resource, view in routes:
            if hasattr(view, "view_class"):
                self.sanic.add_route(view, resource)
            else:
                self.sanic.add_route(view.as_view(), resource)

    def run(self, host="0.0.0.0", **kwargs):
        """Run sanic application."""
        app = AppLoader(factory=lambda: self.sanic).load()

        sock = kwargs.pop("sock", None)
        if sock:
            _, port = sock.getsockname()
            app.prepare(host=host, port=port, **kwargs)
            # if sock is specified sanic wont bind a new one
            # otherwise app will get stuck and tests wont pass
            app.state.server_info[0].settings["sock"] = sock
        else:
            port = kwargs.pop("port", None) or self.cmdArgs.port
            app.prepare(host=host, port=port, **kwargs)
        Sanic.serve_single(primary=app)


class LunaErrorHandler(ErrorHandler):
    """Common error handler"""

    def default(self, request: VLRequest, exception):
        """Default error handler."""
        return serverErrorHandler(request, exception)


class Sanic(BaseSanic):  # pylint: disable-msg=too-many-ancestors
    """Sanic customization."""

    async def _static_request_handler(self, *args, **kwargs):  # pylint: disable-msg=C0103,W0221
        """
        Overrides sanic static processing because it returns None on `IsADirectory` exception that leads to 500 error
        """
        result = await super()._static_request_handler(*args, **kwargs)  # pylint: disable-msg=E1101
        if result is None:
            raise VLException(Error.PageNotFoundError, 404, False)
        return result

    async def _websocket_handler(
        self, handler: "WSBaseHandler", request: VLRequest, *args, subprotocols=None, **kwargs
    ):
        """
        Websocket handle. Open ws session with validation and preparation.
        """
        wsSession = getattr(handler, "wsSession", None)  # hack for data validation before handshake
        if wsSession:
            handler = handler(request, wsSession)
            async with handler.wsSession(request):
                await super()._websocket_handler(
                    lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs
                )
        else:
            await super()._websocket_handler(
                lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs
            )

    def stop(self, *args, **kwargs):
        """
        Stops sanic application, taking care of not shutting down twice.
        """
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        super().stop(*args, **kwargs)
